# -*- coding: utf-8 -*-
import inspect
import json
import logging
import threading
import time
from abc import ABC
from importlib import import_module

from django.conf import settings
from django.core import management
from django.core.serializers import serialize
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.db.models import Q
from django.utils import module_loading, timezone


class JobHandler:
    send_notifications = True
    serializer_method = str
    serializer_params = {}

    def __init__(self, instance, **kwargs):
        self.parameters = kwargs
        self.instance = instance
        self.logs_cleaning()
        self.logger = logging.getLogger('cron')
        self.logger.info("***************************************")
        self.logger.info('LOG' + " " + self.instance.name + " id: " + str(self.instance.pk))
        self.logger.info('RECEIVED PARAMETERS' + str(self.parameters))
        self.logger.info("***************************************")
        self.elaboration_time = None
        self.elaboration_results = None
        self.is_valid = None
        self.is_deletion_valid = None
        self.deletion_results = None
        self.extra_info = None
        self.log_file_content = None

    def do(self):
        start = timezone.now()
        try:
            self.elaboration_results = self.get_elaboration_results(self.run())
            self.is_valid = self.get_is_valid()
        except Exception as e:
            self.elaboration_results = {
                "error_message": str(e),
                "error_type": e.__class__.__name__,
            }
            self.is_valid = False
        end = timezone.now()
        self.elaboration_time = end - start
        return self

    def delete(self):
        try:
            self.deletion_results = self.get_elaboration_results(self.remove())
            self.is_deletion_valid = True
        except Exception as e:
            self.deletion_results = {
                "error_message": str(e),
                "error_type": e.__class__.__name__,
            }
            self.is_deletion_valid = False
        return self

    def logs_cleaning(self):
        """Delete old job logs"""
        now, query = timezone.now(), Q()
        success = settings.CRON_JOBLOG_SUCCESS_DURATION
        fail = settings.CRON_JOBLOG_FAIL_DURATION

        if not success and not fail:
            return

        if success:
            query |= Q(is_valid=True, date_last_modify__lt=now - success)
        if fail:
            query |= Q(is_valid=False, date_last_modify__lt=now - fail)

        self.instance.job_elaborations.filter(query).delete()

    def get_elaboration_results(self, run_results: dict):
        serializer_method = getattr(self, 'serializer_method')
        return serializer_method(run_results, **self.serializer_params)

    def get_is_valid(self) -> bool:
        """Return True if elaboration results are valid."""
        raise NotImplementedError("get_is_valid() must be implemented")

    def run(self):
        """
        Write Job logic here.
        self.parameters contains the required parameters.
        """
        raise NotImplementedError("run() must be implemented")

    def remove(self):
        """
        Override this method for custom remove Job logic.
        self.parameters contains the required parameters.
        """
        self.instance.job_elaborations.all().delete()
        return True, "JobLog deleted"

    @classmethod
    def create(cls, creator, job_name, job_command, parameters={}):
        from thx.cron.models import Job
        try:
            job = Job.objects.create(
                name=job_name,
                command=job_command,
                is_una_tantum=True,
                creator=creator,
                last_modifier=creator,
                parameters=parameters,
            )
            return True, job
        except Exception as e:
            return False, e


class JSONStringJobHandler(JobHandler, ABC):
    serializer_method = staticmethod(json.dumps)
    serializer_params = {'cls': DjangoJSONEncoder}


class JSONJobHandler(JobHandler, ABC):
    def get_elaboration_results(self, run_results: dict):
        return run_results


class ModelSerializeDjangoJSONEncoder(DjangoJSONEncoder):
    def default(self, o):
        if isinstance(o, models.Model):
            return serialize('json', [o])
        return super().default(o)


class JSONModelSerializeJobHandler(JobHandler, ABC):
    serializer_method = staticmethod(json.dumps)
    serializer_params = {'cls': ModelSerializeDjangoJSONEncoder}


JOB_HANDLER_CLASSES = [
    JobHandler.__name__,
    JSONStringJobHandler.__name__,
    JSONJobHandler.__name__,
    ModelSerializeDjangoJSONEncoder.__name__,
] + list(settings.THX_CRON_JOB_HANDLER_CLASSES)


def get_command():
    from django.apps import apps

    commands_list = []
    app_list = apps.get_app_configs()
    for app in app_list:
        try:
            import_module(f'{app.name}.jobs')
            jobs_module = module_loading.import_string(f'{app.name}.jobs')
            for name in dir(jobs_module):
                obj = getattr(jobs_module, name)
                if inspect.isclass(obj) and issubclass(obj, JobHandler) and obj.__name__ not in JOB_HANDLER_CLASSES:
                    commands_list.append(app.name + ".jobs." + obj.__name__)
        except ModuleNotFoundError:
            pass

    return [(command, command) for command in commands_list]


class AsyncRunJobs(threading.Thread):
    def run(self, sleep: int = 0):
        time.sleep(sleep)
        self._run_jobs()

    @staticmethod
    def _run_jobs():
        management.call_command('run_jobs')
